﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Data.Objects.DataClasses;
using System.Data.Objects;
using System.Data;
using System.Runtime.Serialization;
using System.Data.Metadata.Edm;
using System.Diagnostics;
using System.Data.EntityClient;
using System.Configuration;

namespace Perseus
{
    // This class is a data contract serializable image of an ObjectContext.  You can construct it
    // given a context, and if you then serialize and deserialize, you can apply the snapshot to
    // a "blank" context that you have constructed.
    //
    // NOTE: If you construct one, and then immediately try to apply it to a context, the apply
    // will fail because the entities will still be attached to the context which the snapshot
    // was originally constructed from.

    [DataContract]
    public class ContextSnapshot
    {
        //
        // serialized members
        //

        [DataMember] string connectionString;

        [DataMember] List<IEntityWithKey> unchangedEntities;
        [DataMember] List<IEntityWithKey> modifiedEntities;
        [DataMember] List<IEntityWithKey> modifiedOriginalEntities;
        [DataMember] List<IEntityWithKey> deletedEntities;
        [DataMember] List<IEntityWithKey> addedEntities;

        [DataMember] List<RelationshipEntry> addedRelationships;
        [DataMember] List<RelationshipEntry> deletedRelationships;
        [DataMember] List<RelationshipEntry> unchangedManyToManyRelationships;

        //
        // non-serialized members
        //

        Dictionary<EntityKey, int> addedEntityKeyToIndex;

        //
        // properties
        //

        public string ConnectionString
        {
            get
            {
                return this.connectionString;
            }
            private set
            {
                // Clear the store provider connection string because it could contain a password
                // or something else we don't want to send to the client.
                var csBuilder = new EntityConnectionStringBuilder(value);
                if ((csBuilder.Name != null) && (csBuilder.Name != ""))
                {
                    // lookup the actual connection string from the config file
                    string connectionString = ConfigurationManager.ConnectionStrings[csBuilder.Name].ConnectionString;
                    csBuilder = new EntityConnectionStringBuilder(connectionString);
                }
                csBuilder.ProviderConnectionString = null;
                this.connectionString = csBuilder.ConnectionString;
            }
        }

        public Dictionary<EntityKey, int> AddedEntityKeyToIndex
        {
            get
            {
                return this.addedEntityKeyToIndex;
            }
        }

        public IEntityWithKey GetAddedEntity(int addedEntityIndex)
        {
            return this.addedEntities[addedEntityIndex];
        }
        
        //
        // constructors
        //

        public ContextSnapshot(ObjectContext context) : this(context, context.Connection.ConnectionString)
        {
        }

        public ContextSnapshot(ObjectContext context, string connectionString)
        {
            this.ConnectionString = connectionString;

            // For unchanged entities we only need the current values.
            this.unchangedEntities = new List<IEntityWithKey>(context.GetEntities(EntityState.Unchanged));

            // For modified entities we need current plus the original values object along with EntityReferences
            // (the addition of references to the original values object here is critical since this is how
            // 1-1 and 1-many relationships are serialized for modified entities since it's the original values
            // object which is attached on modified entities.  In the case of deleted entities, the relationships
            // don't exist, and in the cases of added or unchanged entities the current values objects bring
            // along the relationships and they are what are added or attached.
            this.modifiedEntities = new List<IEntityWithKey>();
            this.modifiedOriginalEntities = new List<IEntityWithKey>();
            foreach (var entity in context.GetEntities(EntityState.Modified))
            {
                this.modifiedOriginalEntities.Add((IEntityWithKey)context.CreateOriginalValuesObjectWithReferences(entity));
                this.modifiedEntities.Add(entity);
            }

            // For deleted, just the original values.
            this.deletedEntities = new List<IEntityWithKey>();
            foreach (var entity in context.GetEntities(EntityState.Deleted))
            {
                this.deletedEntities.Add((IEntityWithKey)context.CreateOriginalValuesObject(entity));
            }

            // For added entities we need the current values, plus we need to make an index mapping from 
            // the key to where the entity lands in the added entities list so that we can serialize those
            // index values where we would have serialized the added entity key.
            this.addedEntities = new List<IEntityWithKey>();
            this.addedEntityKeyToIndex = new Dictionary<EntityKey, int>();
            int index = 0;
            foreach (var entity in context.GetEntities(EntityState.Added))
            {
                this.addedEntities.Add(entity);
                this.addedEntityKeyToIndex.Add(entity.EntityKey, index);
                index++;
            }

            // Also serialize lists of added, deleted relationships.
            this.addedRelationships = new List<RelationshipEntry>();
            foreach (var stateEntry in context.GetRelationships(EntityState.Added))
            {
                this.addedRelationships.Add(new RelationshipEntry(stateEntry, context, this.addedEntityKeyToIndex));
            }
            this.deletedRelationships = new List<RelationshipEntry>();
            foreach (var stateEntry in context.GetRelationships(EntityState.Deleted))
            {
                this.deletedRelationships.Add(new RelationshipEntry(stateEntry, context, this.addedEntityKeyToIndex));
            }

            // Serialize unchanged many-to-many relationships
            this.unchangedManyToManyRelationships = new List<RelationshipEntry>();
            foreach (var stateEntry in context.GetUnchangedManyToManyRelationships())
            {
                this.unchangedManyToManyRelationships.Add(new RelationshipEntry(stateEntry, context, this.AddedEntityKeyToIndex));
            }
        }

        public void ApplyToContext(ObjectContext context)
        {
            // Attach all the unchanged entities.
            foreach (IEntityWithKey entity in this.unchangedEntities)
            {
                context.Attach(entity);
            }

            // Attach & apply changes to modified entities
            foreach (IEntityWithKey entity in this.modifiedOriginalEntities)
            {
                context.Attach(entity);
            }
            foreach (IEntityWithKey entity in this.modifiedEntities)
            {
                context.ApplyPropertyChanges(entity.EntityKey.EntityContainerName + "." +
                        entity.EntityKey.EntitySetName, entity);
            }

            // Add entities & relationships
            foreach (IEntityWithKey entity in this.addedEntities)
            {
                // have to remove the entity key from the object so add will succeed
                string fullEntitySetName = entity.EntityKey.GetFullEntitySetName();
                entity.EntityKey = null;
                context.AddObject(fullEntitySetName, entity);
            }
            foreach (RelationshipEntry re in this.addedRelationships)
            {
                re.AddRelationship(context, this.addedEntities);
            }

            // to deal with deleting entities & relationships properly, we need a special dance:
            //  - attach all the deleted entities & the deleted relationships
            //  - delete the entities
            //  - delete the relationships
            foreach (IEntityWithKey entity in this.deletedEntities)
            {
                context.Attach(entity);
            }
            foreach (RelationshipEntry re in this.deletedRelationships)
            {
                re.AttachRelationship(context);
            }
            foreach (IEntityWithKey entity in this.deletedEntities)
            {
                context.DeleteObject(entity);
            }
            foreach (RelationshipEntry re in this.deletedRelationships)
            {
                re.DeleteRelationship(context);
            }
            foreach (RelationshipEntry re in this.unchangedManyToManyRelationships)
            {
                re.AttachRelationship(context);
            }

            // Clear lists.  Once a snapshot has been applied, you cannot re-apply it
            // since the entities have been attached.
            this.unchangedEntities = null;
            this.modifiedOriginalEntities = null;
            this.modifiedEntities = null;
            this.deletedEntities = null;
            this.addedEntities = null;
        }
    }
}
