﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Data.Objects.DataClasses;
using System.Data.Objects;
using System.Data;
using System.Runtime.Serialization;
using System.Data.Metadata.Edm;
using System.Diagnostics;
using System.Data.EntityClient;

// Perseus is a small project designed to explore ways for exchanging graphs of entities 
// over WCF web services.  The key piece of the project is EntityBag<T> which stores a
// graph of entities along with changing tracking information.  Here's hoping no one will
// use this to store & transport something as nasty as Medusa's head.  ;-)

// Sample usage:
//      Service method for retrieval:
//          public EntityBag<Customer> GetCustomerAndOrders()
//          {
//              using (var ctx = new MyContext())
//              {
//                  var customer = ctx.Customers.Include("Orders").First();
//                  return ctx.CreateEntityBag<Customer>(customer);
//              }
//          }
//      Service method for update:
//          public void UpdateCustomer(EntityBag<Customer> bag)
//          {
//              using (var ctx = new MyContext())
//              {
//                  bag.UnwrapInto(ctx);
//                  ctx.SaveChanges();
//              }
//          }
//      Client code:
//          using (ServiceClient client = new ServiceClient())
//          {
//              var bag = client.GetCustomerAndOrders();
//              bag.root.Name = "foo";  // modifiy a property
//              bag.root.Orders.Add(new Order("bar"));  // add something new to the graph
//              // do other modifications...
//              client.UpdateCustomer(bag);
//          }

// Simplifying assumptions which could be removed later:
//      - We only handle entities which implement IEntityWithKey. Post beta 3 we can get
//        back a method that will look up an object in the state manager without requiring
//        that it implement IEntityWithKey or that the calling code know the entityset,
//        but for now we'll leave as is.
//      - New EntityBags are constructed only with entities that are attached to a context.

namespace Perseus
{
    public enum Mode
    {
        Constructed,
        Deserialized,
        LocalContext,
        Unwrapped
    }

    public static class EntityBagExtensionMethods
    {
        // This public extension method is what you use to create an EntityBag:
        //      var bag = ctx.CreateEntityBag<MyType>(myObject);
        public static EntityBag<T> CreateEntityBag<T>(this ObjectContext context, T root) where T : IEntityWithKey
        {
            return new EntityBag<T>(context, root);
        }
    }

    [DataContract]
    public class EntityBag<T> where T : IEntityWithKey
    {
        //
        // serialized members
        //

        [DataMember] ContextSnapshot snapshot;

        // We store the EntityKey of the root entity so that we can expose it as an entry point.
        // Since entities in the added state have a tempkey (which doesn't serialize effectively)
        // if the root entity is in the added state, we store the index of it in the list of
        // added entities so that we can determine the new temp key after deserialization.
        [DataMember] EntityKey rootKey;
        [DataMember] int rootAddedEntityIndex;

        //
        // non-serialized members
        //

        ObjectContext context;
        string connectionString;
        
        //
        // properties
        //

        public Mode Mode { get; private set; }
        
        public T Root
        {
            get
            {
                if (this.Mode == Mode.Deserialized)
                {
                    SwitchToLocalContext();
                }

                return (T) context.GetEntityByKey(this.rootKey);
            }
            internal set
            {
                this.rootKey = value.EntityKey;
            }
        }

        //
        // methods
        //

        // The only public way to construct an entity bag is by using the factory method 
        // which is an extension to ObjectContext -- this simplifies some scenarios.
        internal EntityBag(ObjectContext sourceContext, T root)
        {
            this.context = sourceContext;
            // NOTE: We need to capture the connection string now so we'll have it when serializing--in many
            // scenarios the connection will be closed and the connection string lost by the time we serialize.
            this.connectionString = sourceContext.Connection.ConnectionString;
            this.Root = root;
            this.Mode = Mode.Constructed;
        }

        [OnSerializing]
        void OnSerializing(StreamingContext streamingContext)
        {
            switch(this.Mode)
            {
                case Mode.Unwrapped:
                    throw new InvalidOperationException("Cannot serialize when Mode == Unwrapped.");
                case Mode.Constructed:
                    this.snapshot = new ContextSnapshot(this.context, this.connectionString);
                    this.PrepareRootEntityForSerialization();
                    break;
                case Mode.LocalContext:
                    this.snapshot = new ContextSnapshot(this.context);
                    this.PrepareRootEntityForSerialization();
                    break;
                default:
                    // Mode.Deserialized -- nothing to do
                    break;
            }
        }

        [OnDeserialized]
        void OnDeserialized(StreamingContext streamingContext)
        {
            this.Mode = Mode.Deserialized;
        }

        public void Delete(object entity)
        {
            switch (this.Mode)
            {
                case Mode.Deserialized:
                    SwitchToLocalContext();
                    break;
                case Mode.LocalContext:
                    break;
                default:
                    throw new InvalidOperationException("Can't delete an object when in Constructed or Unwrapped mode.");
            }
            this.context.DeleteObject(entity);
        }

        public void UnwrapInto(ObjectContext context)
        {
            switch(this.Mode)
            {
                case Mode.Deserialized:
                    // This is the normal case for unwrap into -- just apply the snapshot
                    this.snapshot.ApplyToContext(context);

                    break;
                case Mode.LocalContext:
                    // If we are operating on the local context, then we need to take a new snapshot,
                    // detach things that will conflict and then apply.
                    this.snapshot = new ContextSnapshot(this.context);

                    // Detach unchanged and added entities from the internal context so the snapshot 
                    // can be applied elsewhere.  Don't have to worry about modified entities because
                    // only the modifiedOriginal versions are attached during the apply, and those
                    // have been synthesized from the original values--they aren't attached.
                    foreach (var entity in this.context.GetEntities(EntityState.Unchanged))
                    {
                        this.context.Detach(entity);
                    }
                    foreach (var entity in this.context.GetEntities(EntityState.Added))
                    {
                        // Detaching an added entity will clear its key -- so save before detaching
                        // and then restore.  The key is needed since it is what stores the entityset.
                        EntityKey key = entity.EntityKey;
                        this.context.Detach(entity);
                        entity.EntityKey = key;
                    }

                    this.snapshot.ApplyToContext(context);
                    this.snapshot = null;
                    break;
                default:
                    // mode == constructed || unwrapped
                    throw new InvalidOperationException("Can only unwrap if Mode == Deserialized || LocalContext.");
            }

            // If we make it this far, then we have successfully unwrapped.
            this.snapshot = null;
            this.Mode = Mode.Unwrapped;            
        }

        void SwitchToLocalContext()
        {
            Debug.Assert(this.Mode == Mode.Deserialized);

            this.context = new ObjectContext(this.snapshot.ConnectionString);
            this.snapshot.ApplyToContext(this.context);

            // Set the root key if root entity was added (in which case the root key is null)
            if (this.rootKey == null)
            {
                this.Root = (T)this.snapshot.GetAddedEntity(this.rootAddedEntityIndex);
            }

            this.Mode = Mode.LocalContext;
        }

        void PrepareRootEntityForSerialization()
        {
            if (this.context.ObjectStateManager.GetObjectStateEntry(this.rootKey).State == EntityState.Added)
            {
                this.rootAddedEntityIndex = this.snapshot.AddedEntityKeyToIndex[this.rootKey];
                this.rootKey = null;
            }
        }
    }
}
